/*
** @file date.c
**
** libdate - A C library for manipulating dates.
** Copyright (C) 2000 James Williams
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "config.h"
#include "date.h"

/** The names of the months */
static char *month[] = {"January", "February", "March", "April", "May",
			"June", "July", "August", "September", "October",
			"November", "December"};

/** The names of all the days of the week */
static char *dow[] = {"Sunday", "Monday", "Tuesday",
		      "Wednesday", "Thursday", "Friday", "Saturday"};

/** The number of days in each month */
static int days_per_month[] = { 31, 28, 31, 30, 31, 30,
				31, 31, 30, 31, 30, 31 };

/** Modulo function that works correctly for negative divisors */
#define modulo(x,y)  ((x)%(y)<0?(x)%(y)+(y):(x)%(y))

/* See if a date occurs within an event's active range */
#define DT_event_is_in_range(d,e)  \
   ((d)->rata_die >= (e)->start_date->rata_die \
      && ((e)->end_date == NULL \
          || (d)->rata_die <= (e)->end_date->rata_die))

/*
** DT_mkdate
**
** Create a DT_DATE structure from a given year, month, and date
**
** Inputs:
**   year: The full year (not just the last two digits)
**   month: The month (1=Jan, 2=Feb, ... , 11=Nov, 12=Dec)
**   date: The day of the month
**
** Returns:
**   A DT_DATE structure with the given date, or NULL if there is an error.
**
** Notes:
**   This function does no sanity checking on the supplied date.  So if
**   you enter 40 for the day, or 20 for the month, it will be accepted.
*/
DT_DATE *DT_mkdate(int year, int month, int day)
{
  DT_DATE *newdate;

  if ((newdate = malloc(sizeof(DT_DATE))) == NULL)
    {
      /* Out of memory */
      return NULL;
    }

  DT_set_date(newdate, year, month, day);
  return newdate;
}

/*
** DT_mkdate_dow
**
** Sets the date according to a particular occurance of a day of the week
** in some month and year.  This can be used to, for example, find the
** first Sunday in April of 2000, or the last day in October of 2000.
**
** Inputs:
**   ordinal: DT_FIRST, DT_SECOND, DT_THIRD, DT_FOURTH, or DT_LAST.
**   dow: Day of week (0-6), or DT_DAY
**   month: The month (1-12)
**   year: The year
**
** Returns:
**   A DT_DATE structure with the given date, or NULL if there is an error.
*/
DT_DATE *DT_mkdate_dow(int ordinal, int dow, int month, int year)
{
  DT_DATE *newdate;

  if ((newdate = malloc(sizeof(DT_DATE))) == NULL)
    {
      /* Out of memory */
      return NULL;
    }

  DT_set_year(newdate, year);
  DT_find_dow_date(newdate, ordinal, dow, month);
  return newdate;
}

/*
** DT_mkdate_rata_die
**
** Given a date in rata die format, it returns a DT_DATE structure for it.
**
** Inputs:
**   rata_die: The date in rata die format
**
** Returns:
**   A DT_DATE structure with the given date, or NULL if there is an error.
*/
DT_DATE *DT_mkdate_rata_die(long int rata_die)
{
  DT_DATE *newdate;

  if ((newdate = malloc(sizeof(DT_DATE))) == NULL)
    {
      /* Out of memory */
      return NULL;
    }

  DT_set_rata_die(newdate, rata_die);
  return newdate;
}

/*
** DT_copy_date
**
** Creates a new date structure by copying the elements from an existing
** one.
**
** Inputs:
**   date: The date to copy
**
** Returns:
**   A pointer to a new date which is equivalent to the input date, or NULL
**   if there is an error.
*/
DT_DATE *DT_copy_date(DT_DATE *date)
{
  DT_DATE *newdate;

  if ((newdate = malloc(sizeof(DT_DATE))) == NULL)
    {
      /* Out of memory */
      return NULL;
    }
  
  memcpy(newdate, date, sizeof(DT_DATE));
  return newdate;
}

/*
** DT_rmdate
**
** Frees a DT_DATE structure.  Call this when you are finished using one.
**
** Inputs:
**   date: The DT_DATE structure to free
**
** Returns:
**   void
*/
void DT_rmdate(DT_DATE *date)
{
  if (date)
    free(date);
}

/*
** DT_days_between
**
** Determines how many days are between two dates.  It subtracts the
** second from the first, so if the second date is more recent than
** the first, the result will be negative.
**
** Inputs:
**   date1: The first date
**   date2: The second date
**
** Returns:
**   The number of days between the two dates, with a negative value
**   if the second is more recent than the first.
*/
long int DT_days_between(DT_DATE *date1, DT_DATE *date2)
{
  return date1->rata_die - date2->rata_die;
}

/*
** DT_months_between
**
** Determines how many months are between two dates.  It subtracts the
** second from the first, so if the second date is more recent than
** the first, the result will be negative.  Also, this function
** completely ignores the day of the month, so consecutive days in different
** months, like Jan 31 and Feb 1 will have a difference of 1.
**
** Inputs:
**   date1: The first date
**   date2: The second date
**
** Returns:
**   The number of months between the two dates, with a negative value
**   if the second is more recent than the first.
*/
long int DT_months_between(DT_DATE *date1, DT_DATE *date2)
{
  return ((DT_year(date1)*12 + DT_month(date1))
	  - (DT_year(date2)*12 + DT_month(date2)));
}

/*
** DT_ascii
**
** Converts a date into ascii format suitable for displaying to the user.
** The format can be in one of several formats, which are DT_US_FORMAT,
** DT_EUROPE_FORMAT, or DT_UNIVERSAL_FORMAT.
** This symbol can be ORed with DT_NUMERIC_OUT if you want the output to
** be strictly numeric.  The result will be stored in a statically
** declared buffer.
**
** Inputs:
**   date: The date to convert
**   format: The desired output format, as described above
**
** Returns:
**   A pointer to a string containing the ASCII version of the date
**
** Notes:
**   This version is not thread-safe
*/
char *DT_ascii(DT_DATE *date, unsigned long int format)
{
  static char buffer[DT_ASCII_BUFFER_SIZE];

  return DT_ascii_r(buffer, DT_ASCII_BUFFER_SIZE, date, format);
}

/*
** DT_ascii_r
**
** Converts a date into ascii format suitable for displaying to the user.
** The format can be in one of several formats, which are DT_US_FORMAT,
** DT_EUROPE_FORMAT, or DT_UNIVERSAL_FORMAT.
** This symbol can be ORed with DT_NUMERIC_OUT if you want the output to
** be strictly numeric.  The result will be stored in a statically
** declared buffer.
**
** Inputs:
**   buffer: The buffer where the output will be held
**   buf_size: The size of the buffer
**   date: The date to convert
**   format: The desired output format, as described above
**
** Returns:
**   A pointer to a string containing the ASCII version of the date
**
** Notes:
**   This version is thread-safe
*/
char *DT_ascii_r(char *buffer, int buf_size, DT_DATE *date,
		 unsigned long int format)
{
  char local_buffer[DT_ASCII_BUFFER_SIZE];

  buffer[0] = '\0';

  if (format & DT_US_FORMAT)
    {
      if (format & DT_NUMERIC_OUT)
	{
	  sprintf(local_buffer, "%02d/%02d/%d",
		  DT_month(date), DT_day(date), DT_year(date));
	}
      else
	{
	  sprintf(local_buffer, "%s %d, %d",
		  month[DT_month(date)-1], DT_day(date), DT_year(date));
	}
    }
  else if (format & DT_EUROPE_FORMAT)
    {
      if (format & DT_NUMERIC_OUT)
	{
	  sprintf(local_buffer, "%02d-%02d-%02d",
		  DT_day(date), DT_month(date), DT_year(date));
	}
      else
	{
	  sprintf(local_buffer, "%d-%s-%d",
		  DT_day(date), month[DT_month(date)-1], DT_year(date));
	}
    }
  else if (format & DT_UNIVERSAL_FORMAT)
    {
      if (format & DT_NUMERIC_OUT)
	{
	  sprintf(local_buffer, "%d%02d%02d",
		  DT_year(date), DT_month(date), DT_day(date));
	}
      else
	{
	  sprintf(local_buffer, "%d %s %02d",
		  DT_year(date), month[DT_month(date)-1], DT_day(date));
	}
    }

  strncpy(buffer, local_buffer, buf_size);
  buffer[buf_size-1] = '\0';

  return buffer;
}

/*
** DT_dow_ascii
**
** Given a date, returns a pointer to an ascii version of the day of the
** week for that date.
**
** Inputs:
**   date: The date to convert
**
** Returns:
**   A pointer to the ascii version of the day of the week
*/
char *DT_dow_ascii(DT_DATE *date)
{
  return (dow[DT_dow(date)]);
}

/*
** DT_set_date
**
** Sets the date in a date structure.
**
** Inputs:
**   date: The DT_DATE structure
**   year: The year
**   month: The month
**   day: The day of the month
**
** Returns:
**   void
*/
void DT_set_date(DT_DATE *date, int year, int month, int day)
{
  long int rd;

  /* Convert the date to Rata Die format */
  year = year + (month-14)/12;

  if (month<3)
    {
      month += 12;
    }

  rd = day
    + (153*month - 457)/5
    + 365*year + year/4 - year/100 + year/400 - 306;

  date->rata_die = rd;
  date->greg_calced = 0;
}

/*
** DT_set_rata_die
**
** Sets the date in a date structure using the rata die format.
**
** Inputs:
**   date: The DT_DATE structure
**   rata_die: The date in rata die format
**
** Returns:
**   void
*/
void DT_set_rata_die(DT_DATE *date, long int rata_die)
{
  date->rata_die = rata_die;
  date->greg_calced = 0;
}

/*
** DT_dow
**
** Returns the day of week for the time stored in a DT_DATE structure.
**
** Inputs:
**   date: The date structure to use
**
** Returns:
**   The day of week (0=Sunday, 1=Monday, ... , 5=Friday, 6=Saturday)
*/
int DT_dow(DT_DATE *date)
{
  return (date->rata_die % 7);
}

/*
** DT_gregorian_date
**
** Internally calculates the gregorian date from the rata die.  This
** function is for internal use only.
**
** Inputs:
**   date: The date structure
**
** Returns:
**   void (All output is stored in date)
*/
static void DT_gregorian_date(DT_DATE *date)
{
  long int z,h,a,b,c;

  if (!date->greg_calced)
    {
      z = date->rata_die+306;
      h = 100*z - 25;
      a = h/3652425;
      b = a - a/4;
      date->year = (100*b + h)/36525;
      c = b + z - 365*date->year - date->year/4;
      date->month = (5*c + 456) / 153;
      date->day = c - (153*date->month - 457)/5;

      if (date->month > 12)
	{
	  date->year++;
	  date->month -= 12;
	}

      date->greg_calced = 1;
    }
}

/*
** DT_year
**
** Returns the year part of the date.
**
** Inputs:
**   date: The date to extract the year from
**
** Returns:
**   The year
*/
int DT_year(DT_DATE *date)
{
  if (!date->greg_calced)  DT_gregorian_date(date);
  return (date->year);
}

/*
** DT_month
**
** Returns the month part of the date.
**
** Inputs:
**   date: The date to extract the month from
**
** Returns:
**   The month (1=Jan ... 12=Dec)
*/
int DT_month(DT_DATE *date)
{
  if (!date->greg_calced)  DT_gregorian_date(date);
  return (date->month);
}

/*
** DT_day
**
** Returns the day part of the date.
**
** Inputs:
**   date: The date to extract the day from
**
** Returns:
**   The day
*/
int DT_day(DT_DATE *date)
{
  if (!date->greg_calced)  DT_gregorian_date(date);
  return (date->day);
}

/*
** DT_rata_die
**
** Returns the date in rata die format
**
** Inputs:
**   date: The date to extract the year from
**
** Returns:
**   The date in rata die format.
*/
long int DT_rata_die(DT_DATE *date)
{
  return date->rata_die;
}

/*
** DT_add_months
**
** Adds some number of months to the date.
**
** Inputs:
**   date: The date to adjust.
**   months: The number of months to add.
**
** Returns:
**   void (All calculations are done inside date)
**
** Notes:
**   If you want to subtract months, pass a negative value in for months.
**   Also, beware of adding months when the day of the month is at the end
**   of the month.  For example, March 31 + 1 month = April 31, but since
**   April only has 30 days, the it rolls back to April 30.
*/
void DT_add_months(DT_DATE *date, long int months)
{
  int year,month,day,maxday;

  year = ((DT_year(date)*12)+DT_month(date)+months)/12;
  month = modulo(DT_month(date)+months, 12);
  maxday = DT_days_this_month_q(year, month);
  day = DT_day(date) > maxday ? maxday : DT_day(date);

  DT_set_date(date, year, month, day);
}

/*
** DT_add_days
**
** Adds some number of days to the date.
**
** Inputs:
**   date: The date to adjust.
**   days: The number of days to add.
**
** Returns:
**   void (All calculations are done inside date)
**
** Notes:
**   If you want to subtract days, pass a negative value in for days.
*/
void DT_add_days(DT_DATE *date, long int days)
{
  date->rata_die += days;
  date->greg_calced = 0;
}

/*
** DT_find_dow_date
**
** Finds a date based on a day of week expression.  For example, to find
** Thanksgiving, one would say the 4th Thursday in November.  The year
** is taken from the current value in date.  It is also possible, using
** DT_DAY for the day of week, to find the last day of the month (or the
** first, second, third, or fourth day of the month, but what's the use
** in that?)
**
** Inputs:
**   date: The DT_DATE structure to modify
**   ordinal: DT_FIRST, DT_SECOND, DT_THIRD, DT_FOURTH, or DT_LAST.
**   dow: Day of week (0-6), or DT_DAY
**   month: The month (1-12)
**
** Returns:
**   void (All results are stored in date)
*/
void DT_find_dow_date(DT_DATE *date, int ordinal, int dow, int month)
{
  int diff;
  int year;

  year = DT_year(date);

  /* Handle the special case of dealing with an arbitrary day */
  if (dow == DT_DAY)
    {
      switch (ordinal)
	{
	  case DT_FIRST:
	  case DT_SECOND:
	  case DT_THIRD:
	  case DT_FOURTH:
	    DT_set_date(date, year, month, ordinal);
	    break;

	  case DT_LAST:
	    DT_set_date(date, year, month+1, 1);
	    DT_add_days(date, -1);
	    break;
	}
    }
  /* Or, find the last occurance of a given weekday in a month */
  else if (ordinal == DT_LAST)
    {
      /* Set the date to the first day of the next month */
      DT_set_date(date, year, month+1, 1);

      /* And adjust the date */
      diff = (DT_dow(date)-dow+7)%7;
      DT_add_days(date, (diff==0?-7:-diff));
    }
  /* Or, find the nth occurance of a weekday in a month */
  else
    {
      /* Set the date to the first day of the month */
      DT_set_date(date, year, month, 1);

      /* Find the requested day */
      DT_add_days(date, (dow-DT_dow(date)+7)%7 + 7*(ordinal-1));
    }
}

/*
** DT_days_this_month
**
** Given a date structure set to some month, return the number of days
** in the month.
**
** Inputs:
**   date: The date whose month we want to check.
**
** Returns:
**   The number of days in the month
*/
int DT_days_this_month(DT_DATE *date)
{
  return (days_per_month[DT_month(date)-1]
	  + ((DT_month(date) == 2 && DT_is_leap_year(date)==1) ? 1 : 0));
}

/*
** DT_days_this_month_q
**
** Given a date structure set to some month, return the number of days
** in the month.  This version is meant to make a quick calculation
** without the need to create a date structure.
**
** Inputs:
**   year: The year for the month we're interested in
**   month: The month we're interested in (1-12)
**
** Returns:
**   The number of days in the month
*/
int DT_days_this_month_q(int year, int month)
{
  return (days_per_month[month-1]
	  + ((month == 2 && DT_is_leap_year_q(year)==1) ? 1 : 0));
}

/*
** DT_easter
**
** Calculate easter for a given year.  The date calculated will be for the
** year contained in the date.
**
** Inputs:
**   date: The DT_DATE structure
**
** Returns:
**   void (date will be updated with easter)
*/
void DT_easter(DT_DATE *date)
{
  int year, month, day, century;
  int G,K,I,J,L;

  year = DT_year(date);
  century = year/100;
  G = year % 19;
  K = (century - 17)/25;
  I = (century - century/4 - (century - K)/3 + 19*G + 15) % 30;
  I = I - (I/28)*(1 - (I/28)*(29/(I + 1))*((21 - G)/11));
  J = (year + year/4 + I + 2 - century + century/4) % 7;
  L = I - J;
  month = 3 + (L + 40)/44;
  day = L + 28 - 31*(month/4);

  DT_set_date(date, year, month, day);
}

/************************* Event Calculations *************************/

/*
** DT_mkevent
**
** Create a basic, no frills event.  The event will recur at regular
** intervals, such as a weekly event (but not a monthly one - see
** DT_mkevent_monthly for that), and have bracketing dates between
** which the event will occur (so you can create an event that will
** only occur for one year, for example).
**
** Inputs:
**   start: The date on which the event first occurs.
**   end: The date after which the event will no longer occur.  If you
**        want the event to continue indefinitely, set this to NULL.
**   interval: How long between one event and the next (in days)?
**
** Returns:
**   A DT_EVENT structure with your event, or NULL on an error.
*/
DT_EVENT *DT_mkevent(DT_DATE *start, DT_DATE *end, long int interval)
{
  DT_EVENT *newevent;

  if ((newevent = (DT_EVENT *)malloc(sizeof(DT_EVENT))) == NULL)
    {
      /* Out of memory */
      return NULL;
    }

  newevent->start_date = start;
  newevent->end_date = end;
  newevent->interval_type = DT_INTERVAL_DAY;
  newevent->interval_length = interval;
  newevent->event_length = 1;
  newevent->adj_rel = 0;
  newevent->adj_sign = 0;
  newevent->pattern = NULL;
  newevent->data = NULL;

  return newevent;
}

/*
** DT_mkevent_monthly
**
** Create an event that will recur on a certain day of the month every
** "interval" months.  You can create a yearly event by setting the
** interval to 12.  The event will first occur on the start day, and
** occur no further than the end date, thus allowing one to bracket
** the dates in which the event will occur.  Set the end date to NULL
** if you want to event to recur indefinetely.
**
** Inputs:
**   start: The date on which the event first occurs.
**   end: The date after which the event will no longer occur.  If you
**        want the event to continue indefinitely, set this to NULL.
**   interval: How long between one event and the next (in months)?
**   day: The day of the month for the event.
**
** Returns:
**   A DT_EVENT structure with your event, or NULL on an error.
*/
DT_EVENT *DT_mkevent_monthly(DT_DATE *start, DT_DATE *end,
			     long int interval, int day)
{
  DT_EVENT *newevent;

  if ((newevent = (DT_EVENT *)malloc(sizeof(DT_EVENT))) == NULL)
    {
      /* Out of memory */
      return NULL;
    }

  newevent->start_date = start;
  newevent->end_date = end;
  newevent->interval_type = DT_INTERVAL_MONTH;
  newevent->interval_length = interval;
  newevent->event_length = 1;
  newevent->day = day;
  newevent->adj_rel = 0;
  newevent->adj_sign = 0;
  newevent->pattern = NULL;
  newevent->data = NULL;

  return newevent;  
}
			  
/*
** DT_mkevent_monthly_dow
**
** Create an event that will recur on a particular day of the week
** at some point in a given month.  This can be used for events which
** occur, for example, on the third Wednesday of every month (or every
** other month, or whatever).  You need to pass a date for the first
** occurance of the event, and optionally a date for the last occurance.
** This last occurance can be NULL if the event to recur indefinitely.
** You must also specify how many months between events.  If you want
** the event to occur every third month, set interval to 3.  The ordinal
** and dow arguments work the same as the ones in DT_find_dow_date.
**
** Inputs:
**   start: The date on which the event first occurs.
**   end: The date after which the event will no longer occur.  If you
**        want the event to continue indefinitely, set this to NULL.
**   interval: How long between one event and the next (in months)?
**   ordinal: DT_FIRST, DT_SECOND, DT_THIRD, DT_FOURTH, or DT_LAST
**   dow: A day of the week, or DT_DAY
**
** Returns:
**   A DT_EVENT structure with your event, or NULL on an error.
*/
DT_EVENT *DT_mkevent_monthly_dow(DT_DATE *start, DT_DATE *end,
				 long int interval, int ordinal, int dow)
{
  DT_EVENT *newevent;

  if ((newevent = (DT_EVENT *)malloc(sizeof(DT_EVENT))) == NULL)
    {
      /* Out of memory */
      return NULL;
    }

  newevent->start_date = start;
  newevent->end_date = end;
  newevent->interval_type = DT_INTERVAL_MONTH_DOW;
  newevent->interval_length = interval;
  newevent->event_length = 1;
  newevent->ordinal = ordinal;
  newevent->day = dow;
  newevent->adj_rel = 0;
  newevent->adj_sign = 0;
  newevent->pattern = NULL;
  newevent->data = NULL;

  return newevent;    
}

/*
** DT_rmevent
**
** Frees the memory associated with an event.  It can optionally free the
** memory associated with the two date structures inside as well.
**
** Inputs:
**   event: The event to free.
**   freeall: Either DT_FREE_ALL or DT_FREE_NONE, referring to whether or
**            not to free the internal date structures.
**
** Returns:
**   void
*/
void DT_rmevent(DT_EVENT *event, int freeall)
{
  if (freeall == DT_FREE_ALL)
    {
      if (event->start_date) DT_rmdate(event->start_date);
      if (event->end_date)   DT_rmdate(event->end_date);
    }

  if (event->pattern)
    {
      free(event->pattern);
    }
  free(event);
}

/*
** DT_event_adj_rel
**
** Give an event a relative adjustment.  What this means is that after
** calculating the day an event occurs on, add this many days to it.  That
** way, you can, for example, calculate election day as the first Tuesday
** after the first Monday of November every fourth year.  Just calculate the
** first Monday in November, then add 1 to it.  That's what this function is
** for.  Notice, however, that the days you add or subtract must not be
** greater than the interval for the event.
**
** Inputs:
**   event: The event to adjust
**   adj: The number of days to add to an event day
**
** Returns:
**   void (the event is updated in place)
*/
void DT_event_adj_rel(DT_EVENT *event, int adj)
{
  event->adj_rel = adj;
}

/*
** DT_event_adj_dow
**
** Given an event, this allows one to adjust to a particular day of the
** week.  The adjustment can be made either in the forward or reverse
** direction, will never exceed one week, and is guaranteed to change
** the date.  As an example, if you wanted the first Monday after April 15,
** and April 15 is on a Friday, this would cause the event to be pushed
** 3 days forward.  If April 15 is on a Monday, the event will be pushed
** 7 days forward.  If you want to disable this feature from an event,
** use DT_NONE for the event, and anything you want for the dow.
**
** Inputs:
**   event: The event to adjust
**   sign: DT_NEXT, DT_PREV, or DT_NONE
**   dow: The day of the week: DT_SUNDAY, DT_MONDAY, etc.
**
** Returns:
**   void (the event is updated in place)
*/
void DT_event_adj_dow(DT_EVENT *event, int sign, int dow)
{
  event->adj_sign = sign;
  event->adj_dow = dow;
}

/*
** DT_event_multiday
**
** Sets an event to last more than a single day.  Using this function,
** every day of an event will be considered part of the event.  For more
** complex patterns, use the DT_event_pattern function instead.
**
** Inputs:
**   event: The event to adjust
**   len: How long to make the event.
**
** Returns:
**   void (the event is updated in place)
**
** Notes:
**   The length should not be longer than the interval between one
**   event and the next.
*/
void DT_event_multiday(DT_EVENT *event, int len)
{
  event->event_length = len;
}

/*
** DT_event_pattern
**
** Sets a complex pattern for a multiday event.  The input contains
** a string of 0's and 1's, with each character representing a single
** day.  A 0 means the event does not take place on that day, while a
** 1 means it does.
**
** Inputs:
**   event: The event to adjust
**   pattern: The pattern for the event
**
** Returns:
**   void (the event is updated in place)
*/
void DT_event_pattern(DT_EVENT *event, char *pattern)
{
  if (event->pattern)
    {
      free(event->pattern);
    }

  if ((event->pattern = strdup(pattern)) == NULL)
    {
      /* Failure!!! */
      event->event_length = 1;
    }
  else
    {
      event->event_length = strlen(pattern);
    }
}

/*
** DT_event_calc_adj
**
** Given a date and an event, adjust the date according to the day and
** day of week adjustments in the event structure.  This assumes the date
** has already been set to match the event sans adjustment. This function is
** for internal use only.
**
** Inputs:
**   date: The date to adjust
**   event: The event from which the adjustments will be taken
**
** Returns:
**   void (The date structure will be adjusted)
*/
static void DT_event_calc_adj(DT_DATE *date, DT_EVENT *event)
{
  /* First, adjust for the +/- days */
  DT_add_days(date, event->adj_rel);

  /* Adjust for the day of week */
  if (event->adj_sign > 0)
    {
      DT_add_days(date, 1);
      DT_add_days(date, (event->adj_dow+7-DT_dow(date))%7);
    }
  else if (event->adj_sign < 0)
    {
      DT_add_days(date, -1);
      DT_add_days(date, -((DT_dow(date)+7-event->adj_dow)%7));
    }
}

/*
** DT_event_check_pattern
**
** Given a date which is assumed to be the beginning of an event, an event
** with a pattern, and a target date, determines if the event occurs on the
** target date.  Notice that this function is for internal use only
**
** Inputs:
**   start_date: The date when the event begins
**   event: The event in question
**   target_date: The target date to test
**
** Returns:
**   1 if the target date is an active event day, or 0 if it's not.
*/
static int DT_event_check_pattern(DT_DATE *start_date, DT_EVENT *event,
				  DT_DATE *target_date)
{
  int offset = DT_days_between(target_date, start_date);

  /* Check the obvious */
  if (event->event_length == 1)
    {
      return (offset==0 ? 1 : 0);
    }

  /* Make sure the target date is within the pattern length */
  if (offset >= event->event_length || offset < 0)
    {
      return 0;
    }

  /* See if the particular date in the pattern is part of the event */
  if (event->pattern)
    {
      return (event->pattern[offset] == '0' ? 0 : 1);
    }
  else
    {
      /* If not pattern is set, assume they are all 1's */
      return 1;
    }
}

/*
** DT_event_calc_from_seed
**
** Given a "seed" date, calculates when the event associated with that
** seed occurs.  Notice that this function is for internal use only.
**
** Inputs:
**   date: The seed date
**   event: The event to calculate
**
** Returns:
**   void (The date is returned in the date structure)
**
** Notes:
**   This function does not take the start and end dates into consideration
**   except to calculate when the interval starts.
*/
static void DT_event_calc_from_seed(DT_DATE *date, DT_EVENT *event)
{
  long int offset;

  switch (event->interval_type)
    {
      case DT_INTERVAL_DAY:
	offset = DT_days_between(date, event->start_date)
	  % event->interval_length;
	DT_add_days(date, -offset);
	break;

      case DT_INTERVAL_MONTH:
	offset = DT_months_between(date, event->start_date)
	  % event->interval_length;
	DT_add_months(date, -offset);
	DT_set_day(date, event->day);
	DT_event_calc_adj(date, event);
	break;

      case DT_INTERVAL_MONTH_DOW:
	offset = DT_months_between(date, event->start_date)
	  % event->interval_length;
	DT_add_months(date, -offset);
	DT_find_dow_date(date,
			 event->ordinal,
			 event->day,
			 DT_month(date));
	DT_event_calc_adj(date, event);
	break;	
    }
}

/*
** DT_event_closest
**
** Find the start of the event on or before a given target date.  This
** function doesn't consider the event's start and end dates.  Notice
** that this function is for internal use only.
**
** Inputs:
**   date: The target date
**   event: The event to find
**
** Returns:
**   void (the date is modified in place)
*/
static void DT_event_closest(DT_DATE *date, DT_EVENT *event)
{
  DT_DATE *event_date;
  DT_DATE *best = NULL;
  int tries = 2;
  int sign;

  /* Make a guess at a seed date */
  event_date = DT_copy_date(date);
  DT_add_days(event_date, -(event->adj_rel));
      
  while (tries--)
    {
      DT_event_calc_from_seed(event_date, event);

      /* Do we have an exact match?  If so, we're done. */
      if (event_date->rata_die == date->rata_die)
	{
	  DT_rmdate(event_date);
	  DT_rmdate(best);
	  return;
	}

      /* Did the estimated date occur before the target date?  If so, it's
       * possibly the correct date */
      if (event_date->rata_die < date->rata_die)
	{
	  DT_rmdate(best);
	  best = DT_copy_date(event_date);
	}
      
      /* Find the next or previous event */
      switch (event->interval_type)
	{
	  case DT_INTERVAL_DAY:
	    sign = event_date->rata_die > date->rata_die ? -1 : 1;
	    DT_add_days(event_date, sign*(event->interval_length));
	    break;

	  case DT_INTERVAL_MONTH:
	  case DT_INTERVAL_MONTH_DOW:
	    sign = event_date->rata_die > date->rata_die ? -1 : 1;
	    DT_set_rata_die(event_date, date->rata_die);
	    DT_add_days(event_date, event->adj_rel);

	    DT_add_months(event_date, sign*(event->interval_length));
	    break;	
	}
    }

  DT_set_rata_die(date, best->rata_die);
  DT_rmdate(event_date);
  DT_rmdate(best);
}

/*
** DT_event_next_block
**
** Given a date which is the start of an event block, find the start of
** the next event block.  Notice that this function is for internal use
** only.
**
** Inputs:
**   event_date: The start of the event block
**   event: The event we are looking at
**
** Returns:
**   void (event_date is modified in place)
*/
static void DT_event_next_block(DT_DATE *event_date, DT_EVENT *event)
{
  switch (event->interval_type)
    {
      case DT_INTERVAL_DAY:
	DT_add_days(event_date, event->interval_length);
	break;

      case DT_INTERVAL_MONTH:
      case DT_INTERVAL_MONTH_DOW:
	DT_add_days(event_date, -(event->adj_rel));
	DT_add_months(event_date, event->interval_length);
	DT_event_calc_from_seed(event_date, event);
	break;
    }
}

/*
** DT_is_event_today
**
** Determine if an event occurs on a particular day.  Given an event and a
** date, determines if the event occurs on that day.
**
** Inputs:
**   date: The DT_DATE structure containing the day
**   event: The DT_EVENT structure containing the event
**
** Returns:
**   1 if the event occurs on the given date, or 0 if it doesn't
*/
int DT_is_event_today(DT_DATE *date, DT_EVENT *event)
{
  DT_DATE *event_date;
  int ret;

  /* See if the date is within the event's bounds */
  if ((event->start_date && date->rata_die < event->start_date->rata_die)
      || (event->end_date && date->rata_die > event->end_date->rata_die))
    {
      return 0;
    }

  /* Find the closest match for the start of an event */
  event_date = DT_copy_date(date);
  DT_event_closest(event_date, event);

  /* See if the date is in the event's pattern */
  ret = DT_event_check_pattern(event_date, event, date);
  DT_rmdate(event_date);
  return ret;
}

/*
** DT_event_next
**
** Given a date and an event, find the next occurance of the event.
** Notice that if an event occurs on the date supplied, that date will
** not be returned.
**
** Inputs:
**   date: The date to check around
**   event: The event to locate
**
** Returns:
**   A pointer to date if a next date was found, or NULL if nothing was
**   found.
**
** Notes:
**   date will be modified in place, and the return value, if successful,
**   will be nothing more than a pointer to it.
*/
DT_DATE *DT_event_next(DT_DATE *date, DT_EVENT *event)
{
  DT_DATE *date1;
  long int offset;

  date1 = DT_copy_date(date);
  DT_add_days(date1, 1);
  DT_event_closest(date1, event);

  /* Determine if the target date occurs within this events pattern */
  if ((offset = DT_days_between(date, date1) + 1) < event->event_length)
    {
      if (event->pattern)
	{
	  /* Scan the pattern for the next occurance */
	  for (; offset < event->event_length; offset++)
	    {
	      if (event->pattern[offset] == '1')
		{
		  if (DT_event_is_in_range(date1,event))
		    {
		      DT_set_rata_die(date, date1->rata_die);
		      DT_add_days(date, offset);
		      return date;
		    }
		  else
		    {
		      return NULL;
		    }
		}
	    }
	}
    }
  
  /* Okay, find the next event block */
  DT_event_next_block(date1, event);
  
  if (event->pattern)
    {
      /* Scan the pattern for the next occurance */
      for (offset = 0; offset < event->event_length; offset++)
	{
	  if (event->pattern[offset] == '1')
	    {
	      if (DT_event_is_in_range(date1,event))
		{
		  DT_set_rata_die(date, date1->rata_die);
		  DT_add_days(date, offset);
		  return date;
		}
	      else
		{
		  return NULL;
		}
	    }
	}
    }
  else
    {
      if (DT_event_is_in_range(date1,event))
	{
	  DT_set_rata_die(date, date1->rata_die);
	  return date;
	}
      else
	{
	  return NULL;
	}
    }

  return NULL;  /* No next event found */
}


/************************* Calendar Rendering *************************/

/*
** DT_cal_rows
**
** Determines how many rows would be in a rendered calendar for a given
** month.
**
** Inputs:
**   date: The DT_DATE structure holding the year and month to check
**
** Returns:
**   The number of rows in a rendered calendar for the month.
*/
int DT_cal_rows(DT_DATE *date)
{
  return (DT_days_this_month(date)-1+(DT_dow(date)+36-DT_day(date))%7)/7+1;
}

/*
** DT_cal_day_to_row
**
** Given a date, return the row in which it would appear in the calendar.
**
** Inputs:
**   date: The date the convert
**
** Returns:
**   The row in which the date appears.
**
** Notes:
**   The first row is row 0.
*/
int DT_cal_day_to_row(DT_DATE *date)
{
  return (DT_day(date)-1+(DT_dow(date)+36-DT_day(date))%7)/7;
}

/*
** DT_cal_day_to_col
**
** Given a date, return the column in which it would appear in the calendar.
**
** Inputs:
**   date: The date the convert
**
** Returns:
**   The column in which the date appears.
**
** Notes:
**   The first column is column 0.
*/
int DT_cal_day_to_col(DT_DATE *date)
{
  /* Duh */
  return DT_dow(date);
}

/*
** DT_cal_row_col_to_day
**
** Given a month, row, and column, determine which day of the month should
** fill that location.
**
** Inputs:
**   date: The DT_DATE structure containing the month and year.
**   row: The row of the calendar (0-5)
**   col: The column of the calendar (0-6)
**
** Returns:
**   The day of the month to fill into the given row and column.
**
** Notes:
**   If this function returns a 0, it means the row and column are for
**   another month.  A -1 means there was an error of some kind.
*/
int DT_cal_row_col_to_day(DT_DATE *date, int row, int col)
{
  int day;

  /* Sanity checks */
  if (row < 0 || row > 5)
    return -1;  /* Out of range */

  if (col < 0 || col > 6)
    return -1;  /* Out of range */

  day = (7*row)+col - (DT_dow(date)+36-DT_day(date))%7 + 1;
  day = day < 1 ? 0 : (day > DT_days_this_month(date) ? 0 : day);

  return day;
}

/************************* Moon Calculations *************************/

/*
** DT_moon_age
**
** Calculate the age of the moon in its synodic month.  A synodic month
** Is the time the moon takes to go though all its phases, from one
** new moon to the next.  Once we know the age, we can infer the phase of
** the moon from it.
**
** Inputs:
**   date: A DT_DATE structure for the date to calculate for.
**   timezone: The timezone, represented as the number of hours relative to GMT
**
** Returns:
**   The age of the moon at midnight on the specified date.
**
** Notes:
**   To find the age of the moon at times other than midnight, just add
**   the fraction of a day corresponding to the time.  In other words,
**   each hour represents 1/24 days, and each minute represents 1/1440
**   days.  If you want 2 30pm (14 30), add 14/24 + 30/1440, (0.604166667).
*/
double DT_moon_age(DT_DATE *date, int timezone)
{
  double cycles;

  cycles = ((double)(date->rata_die)+DT_NEWMOON_OFFSET-((double)timezone/24.0))
    / DT_LUNAR_PERIOD;
  return ((cycles-(int)cycles) * DT_LUNAR_PERIOD);
}
