/*
** date.h
**
** libdate - A C library for manipulating dates
** Copyright (C) 2000 James Williams
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Lesser General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public
** License along with this library; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
*/

#ifndef DATE_H
#define DATE_H

#define DT_INTERVAL_DAY        1
#define DT_INTERVAL_MONTH      2
#define DT_INTERVAL_MONTH_DOW  3

#define DT_FIRST      1
#define DT_SECOND     2
#define DT_THIRD      3
#define DT_FOURTH     4
#define DT_LAST       5

#define DT_SUNDAY     0
#define DT_MONDAY     1
#define DT_TUESDAY    2
#define DT_WEDNESDAY  3
#define DT_THURSDAY   4
#define DT_FRIDAY     5
#define DT_SATURDAY   6
#define DT_DAY        10

#define DT_JANUARY    1
#define DT_FEBRUARY   2
#define DT_MARCH      3
#define DT_APRIL      4
#define DT_MAY        5
#define DT_JUNE       6
#define DT_JULY       7
#define DT_AUGUST     8
#define DT_SEPTEMBER  9
#define DT_OCTOBER    10
#define DT_NOVEMBER   11
#define DT_DECEMBER   12

/* For calculating the phase of the moon */
#define DT_LUNAR_PERIOD    29.5305882
#define DT_NEWMOON_OFFSET  17.7800836001

/* For ASCII output of dates */
#define DT_ASCII_BUFFER_SIZE  32
#define DT_NUMERIC_OUT        0x01
#define DT_US_FORMAT          0x02
#define DT_EUROPE_FORMAT      0x04
#define DT_UNIVERSAL_FORMAT   0x08

/* How to free an event structure */
#define DT_FREE_ALL           1
#define DT_FREE_NONE          2

/* Day of week adjustment sign modifiers */
#define DT_PREV              -1
#define DT_NONE               0
#define DT_NEXT               1

/**
 * The DT_DATE structure is used to store a single date.
 *
 * A date is stored internally in rata die format.  It will also store
 * dates in gregorian format, but the calculation is only done when needed.
 * The greg_calced flag will tell if this operation has been performed for
 * a given rata die date, and if not, it will be, with the results stored
 * in the year, month, and day slots.
 *
 * Do not access this structure directly.  Instead, use the functions
 * available in the date library.
 */ 
typedef struct
{
  long int rata_die;  /**< The date, in rata die format */
  int greg_calced;  /**< A flag which tells in the gregorian date has been calculated */
  int year;  /**< The corresponding year */
  int month;  /**< The corresponding month */
  int day;  /**< The corresponding day of the month */
} DT_DATE;

/**
 * The DT_EVENT structure is used to schedule an event.
 *
 * The event can be a one time event, or a recurring event.  Recurring events
 * can be set up to recur at regular intervals, or at certain times of the
 * month, or even follow a predetermined pattern.
 *
 * Do not access this structure directly.  Instead, use the functions
 * available in the date library.
 */
typedef struct
{
  DT_DATE *start_date;  /**< When the first event is */
  DT_DATE *end_date; /**< Point after which no more events will occur */
  int interval_type; /**< DT_INTERVAL_DAY, DT_INTERVAL_MONTH, DT_INTERVAL_MONTH_DOW */
  int interval_length; /**< Time between events, either in days or months */
  int event_length;  /**< How many days will this event last? */
  int ordinal; /**< DT_FIRST, etc, for weekdays */
  int day; /**< For the specific day of month, or weekday */
  int adj_rel; /**< The number of days to add to the final date */
  int adj_sign; /**< Adjust forward or backward (or not at all) for adj_dow? */
  int adj_dow; /**< Which day to adjust to */
  char *pattern; /**< Irregular event patterns */
  void *data;  /**< User supplied data for this event */
} DT_EVENT;

/* Macros */

/** Determines if a date is in a leap year. */
#define DT_is_leap_year(x)  (DT_year(x)%400==0 \
                             || (DT_year(x)%100!=0 && DT_year(x)%4==0))

/** Determines if a year is a leap year. */
#define DT_is_leap_year_q(x)  ((x)%400==0 || ((x)%100!=0 && (x)%4==0))

/** Sets just the year part of a date. */
#define DT_set_year(date,year)  DT_set_date((date),(year),DT_month(date), \
                                DT_day(date))

/** Sets just the month part of a date. */
#define DT_set_month(date,month)  DT_set_date((date),DT_year(date), \
                                  (month), DT_day(date))

/** Sets just the day part of a date. */
#define DT_set_day(date,day)  DT_set_date((date),DT_year(date), \
                              DT_month(date), (day))

DT_DATE *DT_mkdate(int year, int month, int day);
DT_DATE *DT_mkdate_dow(int ordinal, int dow, int month, int year);
DT_DATE *DT_mkdate_rata_die(long int rata_die);
void DT_rmdate(DT_DATE *date);
DT_DATE *DT_copy_date(DT_DATE *date);
long int DT_days_between(DT_DATE *date1, DT_DATE *date2);
long int DT_months_between(DT_DATE *date1, DT_DATE *date2);
char *DT_ascii(DT_DATE *date, unsigned long int format);
char *DT_ascii_r(char *buffer, int buf_size, DT_DATE *date,
		 unsigned long int format);
char *DT_dow_ascii(DT_DATE *date);
void DT_set_date(DT_DATE *date, int year, int month, int day);
void DT_set_rata_die(DT_DATE *date, long int rata_die);
int DT_dow(DT_DATE *date);
int DT_year(DT_DATE *date);
int DT_month(DT_DATE *date);
int DT_day(DT_DATE *date);
long int DT_rata_die(DT_DATE *date);
void DT_add_days(DT_DATE *date, long int days);
void DT_add_months(DT_DATE *date, long int months);
void DT_find_dow_date(DT_DATE *date, int ordinal, int dow, int month);
int DT_days_this_month(DT_DATE *date);
int DT_days_this_month_q(int year, int month);
void DT_easter(DT_DATE *date);

DT_EVENT *DT_mkevent(DT_DATE *start, DT_DATE *end, long int interval);
DT_EVENT *DT_mkevent_monthly(DT_DATE *start, DT_DATE *end,
			     long int interval, int day);
DT_EVENT *DT_mkevent_monthly_dow(DT_DATE *start, DT_DATE *end,
				 long int interval, int ordinal, int dow);
void DT_rmevent(DT_EVENT *event, int freeall);
void DT_event_adj_rel(DT_EVENT *event, int adj);
void DT_event_adj_dow(DT_EVENT *event, int sign, int dow);
void DT_event_multiday(DT_EVENT *event, int len);
void DT_event_pattern(DT_EVENT *event, char *pattern);
int DT_is_event_today(DT_DATE *date, DT_EVENT *event);
DT_DATE *DT_event_next(DT_DATE *date, DT_EVENT *event);

int DT_cal_rows(DT_DATE *date);
int DT_cal_day_to_row(DT_DATE *date);
int DT_cal_day_to_col(DT_DATE *date);
int DT_cal_row_col_to_day(DT_DATE *date, int row, int col);

double DT_moon_age(DT_DATE *date, int timezone);
#endif  /* DATE_H */
